@@ -652,9 +652,6 @@ int main (int argc, char **argv) {
 
   conf->set("log.thread.name", "true", errstr);
 
-  /* correct producer offsets */
-  conf->set("produce.offset.report", "true", errstr);
-
   /* auto commit is explicitly enabled with --enable-autocommit */
   conf->set("enable.auto.commit", "false", errstr);
 
@@ -665,6 +662,8 @@ int main (int argc, char **argv) {
   conf->set("fetch.wait.max.ms", "500", errstr);
   conf->set("fetch.min.bytes", "4096", errstr);
 
+  conf->set("enable.partition.eof", "true", errstr);
+
   for (int i = 1 ; i < argc ; i++) {
     const char *name = argv[i];
     const char *val = i+1 < argc ? argv[i+1] : NULL;
