@@ -39,8 +39,13 @@
 #include <csignal>
 #include <cstring>
 
+#ifdef _MSC_VER
+#include "../win32/wingetopt.h"
+#elif _AIX
+#include <unistd.h>
+#else
 #include <getopt.h>
-
+#endif
 
 /*
  * Typically include path in a real application would be
@@ -49,8 +54,70 @@
 #include "rdkafkacpp.h"
 
 
+static void metadata_print (const std::string &topic,
+                            const RdKafka::Metadata *metadata) {
+  std::cout << "Metadata for " << (topic.empty() ? "" : "all topics")
+           << "(from broker "  << metadata->orig_broker_id()
+           << ":" << metadata->orig_broker_name() << std::endl;
+
+  /* Iterate brokers */
+  std::cout << " " << metadata->brokers()->size() << " brokers:" << std::endl;
+  RdKafka::Metadata::BrokerMetadataIterator ib;
+  for (ib = metadata->brokers()->begin();
+       ib != metadata->brokers()->end();
+       ++ib) {
+    std::cout << "  broker " << (*ib)->id() << " at "
+              << (*ib)->host() << ":" << (*ib)->port() << std::endl;
+  }
+  /* Iterate topics */
+  std::cout << metadata->topics()->size() << " topics:" << std::endl;
+  RdKafka::Metadata::TopicMetadataIterator it;
+  for (it = metadata->topics()->begin();
+       it != metadata->topics()->end();
+       ++it) {
+    std::cout << "  topic \""<< (*it)->topic() << "\" with "
+              << (*it)->partitions()->size() << " partitions:";
+
+    if ((*it)->err() != RdKafka::ERR_NO_ERROR) {
+      std::cout << " " << err2str((*it)->err());
+      if ((*it)->err() == RdKafka::ERR_LEADER_NOT_AVAILABLE)
+        std::cout << " (try again)";
+    }
+    std::cout << std::endl;
+
+    /* Iterate topic's partitions */
+    RdKafka::TopicMetadata::PartitionMetadataIterator ip;
+    for (ip = (*it)->partitions()->begin();
+         ip != (*it)->partitions()->end();
+         ++ip) {
+      std::cout << "    partition " << (*ip)->id()
+                << ", leader " << (*ip)->leader()
+                << ", replicas: ";
+
+      /* Iterate partition's replicas */
+      RdKafka::PartitionMetadata::ReplicasIterator ir;
+      for (ir = (*ip)->replicas()->begin();
+           ir != (*ip)->replicas()->end();
+           ++ir) {
+        std::cout << (ir == (*ip)->replicas()->begin() ? "":",") << *ir;
+      }
+
+      /* Iterate partition's ISRs */
+      std::cout << ", isrs: ";
+      RdKafka::PartitionMetadata::ISRSIterator iis;
+      for (iis = (*ip)->isrs()->begin(); iis != (*ip)->isrs()->end() ; ++iis)
+        std::cout << (iis == (*ip)->isrs()->begin() ? "":",") << *iis;
+
+      if ((*ip)->err() != RdKafka::ERR_NO_ERROR)
+        std::cout << ", " << RdKafka::err2str((*ip)->err()) << std::endl;
+      else
+        std::cout << std::endl;
+    }
+  }
+}
 
 static bool run = true;
+static bool exit_eof = false;
 
 static void sigterm (int sig) {
   run = false;
@@ -60,8 +127,26 @@ static void sigterm (int sig) {
 class ExampleDeliveryReportCb : public RdKafka::DeliveryReportCb {
  public:
   void dr_cb (RdKafka::Message &message) {
+    std::string status_name;
+    switch (message.status())
+      {
+      case RdKafka::Message::MSG_STATUS_NOT_PERSISTED:
+        status_name = "NotPersisted";
+        break;
+      case RdKafka::Message::MSG_STATUS_POSSIBLY_PERSISTED:
+        status_name = "PossiblyPersisted";
+        break;
+      case RdKafka::Message::MSG_STATUS_PERSISTED:
+        status_name = "Persisted";
+        break;
+      default:
+        status_name = "Unknown?";
+        break;
+      }
     std::cout << "Message delivery for (" << message.len() << " bytes): " <<
-        message.errstr() << std::endl;
+      status_name << ": " << message.errstr() << std::endl;
+    if (message.key())
+      std::cout << "Key: " << *(message.key()) << ";" << std::endl;
   }
 };
 
@@ -72,10 +157,12 @@ class ExampleEventCb : public RdKafka::EventCb {
     switch (event.type())
     {
       case RdKafka::Event::EVENT_ERROR:
+        if (event.fatal()) {
+          std::cerr << "FATAL ";
+          run = false;
+        }
         std::cerr << "ERROR (" << RdKafka::err2str(event.err()) << "): " <<
             event.str() << std::endl;
-        if (event.err() == RdKafka::ERR__ALL_BROKERS_DOWN)
-          run = false;
         break;
 
       case RdKafka::Event::EVENT_STATS:
@@ -97,9 +184,6 @@ class ExampleEventCb : public RdKafka::EventCb {
 };
 
 
-
-
-
 /* Use of this partitioner is pretty pointless since no key is provided
  * in the produce() call. */
 class MyHashPartitionerCb : public RdKafka::PartitionerCb {
@@ -118,6 +202,66 @@ class MyHashPartitionerCb : public RdKafka::PartitionerCb {
   }
 };
 
+void msg_consume(RdKafka::Message* message, void* opaque) {
+  const RdKafka::Headers *headers;
+
+  switch (message->err()) {
+    case RdKafka::ERR__TIMED_OUT:
+      break;
+
+    case RdKafka::ERR_NO_ERROR:
+      /* Real message */
+      std::cout << "Read msg at offset " << message->offset() << std::endl;
+      if (message->key()) {
+        std::cout << "Key: " << *message->key() << std::endl;
+      }
+      headers = message->headers();
+      if (headers) {
+        std::vector<RdKafka::Headers::Header> hdrs = headers->get_all();
+        for (size_t i = 0 ; i < hdrs.size() ; i++) {
+          const RdKafka::Headers::Header hdr = hdrs[i];
+
+          if (hdr.value() != NULL)
+            printf(" Header: %s = \"%.*s\"\n",
+                   hdr.key().c_str(),
+                   (int)hdr.value_size(), (const char *)hdr.value());
+          else
+            printf(" Header:  %s = NULL\n", hdr.key().c_str());
+        }
+      }
+      printf("%.*s\n",
+        static_cast<int>(message->len()),
+        static_cast<const char *>(message->payload()));
+      break;
+
+    case RdKafka::ERR__PARTITION_EOF:
+      /* Last message */
+      if (exit_eof) {
+        run = false;
+      }
+      break;
+
+    case RdKafka::ERR__UNKNOWN_TOPIC:
+    case RdKafka::ERR__UNKNOWN_PARTITION:
+      std::cerr << "Consume failed: " << message->errstr() << std::endl;
+      run = false;
+      break;
+
+    default:
+      /* Errors */
+      std::cerr << "Consume failed: " << message->errstr() << std::endl;
+      run = false;
+  }
+}
+
+
+class ExampleConsumeCb : public RdKafka::ConsumeCb {
+ public:
+  void consume_cb (RdKafka::Message &msg, void *opaque) {
+    msg_consume(&msg, opaque);
+  }
+};
+
 
 
 int main (int argc, char **argv) {
@@ -128,10 +272,10 @@ int main (int argc, char **argv) {
   std::string debug;
   int32_t partition = RdKafka::Topic::PARTITION_UA;
   int64_t start_offset = RdKafka::Topic::OFFSET_BEGINNING;
-  bool exit_eof = false;
   bool do_conf_dump = false;
-  char opt;
+  int opt;
   MyHashPartitionerCb hash_partitioner;
+  int use_ccb = 0;
 
   /*
    * Create configuration objects
@@ -140,10 +284,11 @@ int main (int argc, char **argv) {
   RdKafka::Conf *tconf = RdKafka::Conf::create(RdKafka::Conf::CONF_TOPIC);
 
 
-  while ((opt = getopt(argc, argv, "PCt:p:b:z:qd:o:eX:AM:")) != -1) {
+  while ((opt = getopt(argc, argv, "PCLt:p:b:z:qd:o:eX:AM:f:")) != -1) {
     switch (opt) {
     case 'P':
     case 'C':
+    case 'L':
       mode = opt;
       break;
     case 't':
@@ -229,21 +374,33 @@ int main (int argc, char **argv) {
       }
       break;
 
+      case 'f':
+        if (!strcmp(optarg, "ccb"))
+          use_ccb = 1;
+        else {
+          std::cerr << "Unknown option: " << optarg << std::endl;
+          exit(1);
+        }
+        break;
+
     default:
       goto usage;
     }
   }
 
-  if (mode.empty() || topic_str.empty() || optind != argc) {
+  if (mode.empty() || (topic_str.empty() && mode != "L") || optind != argc) {
   usage:
+	  std::string features;
+	  conf->get("builtin.features", features);
     fprintf(stderr,
             "Usage: %s [-C|-P] -t <topic> "
             "[-p <partition>] [-b <host1:port1,host2:port2,..>]\n"
             "\n"
-            "librdkafka version %s (0x%08x)\n"
+            "librdkafka version %s (0x%08x, builtin.features \"%s\")\n"
             "\n"
             " Options:\n"
             "  -C | -P         Consumer or Producer mode\n"
+            "  -L              Metadata list mode\n"
             "  -t <topic>      Topic to fetch / produce\n"
             "  -p <num>        Partition (random partitioner)\n"
             "  -p <func>       Use partitioner:\n"
@@ -263,6 +420,8 @@ int main (int argc, char **argv) {
             "will be set on topic object.\n"
             "                  Use '-X list' to see the full list\n"
             "                  of supported properties.\n"
+            "  -f <flag>       Set option:\n"
+            "                     ccb - use consume_callback\n"
             "\n"
             " In Consumer mode:\n"
             "  writes fetched messages to stdout\n"
@@ -273,8 +432,9 @@ int main (int argc, char **argv) {
             "\n",
 	    argv[0],
 	    RdKafka::version_str().c_str(), RdKafka::version(),
-	    RdKafka::Conf::DEBUG_CONTEXTS.c_str());
-    exit(1);
+		features.c_str(),
+	    RdKafka::get_debug_contexts().c_str());
+	exit(1);
   }
 
 
@@ -293,7 +453,6 @@ int main (int argc, char **argv) {
   ExampleEventCb ex_event_cb;
   conf->set("event_cb", &ex_event_cb, errstr);
 
-
   if (do_conf_dump) {
     int pass;
 
@@ -327,11 +486,17 @@ int main (int argc, char **argv) {
     /*
      * Producer mode
      */
+
+    if(topic_str.empty())
+      goto usage;
+
     ExampleDeliveryReportCb ex_dr_cb;
 
     /* Set delivery report callback */
     conf->set("dr_cb", &ex_dr_cb, errstr);
 
+    conf->set("default_topic_conf", tconf, errstr);
+
     /*
      * Create producer using accumulated global configuration.
      */
@@ -343,58 +508,69 @@ int main (int argc, char **argv) {
 
     std::cout << "% Created producer " << producer->name() << std::endl;
 
-    /*
-     * Create topic handle.
-     */
-    RdKafka::Topic *topic = RdKafka::Topic::create(producer, topic_str,
-						   tconf, errstr);
-    if (!topic) {
-      std::cerr << "Failed to create topic: " << errstr << std::endl;
-      exit(1);
-    }
 
     /*
      * Read messages from stdin and produce to broker.
      */
-    for (std::string line; run and std::getline(std::cin, line);) {
+    for (std::string line; run && std::getline(std::cin, line);) {
       if (line.empty()) {
         producer->poll(0);
 	continue;
       }
 
+      RdKafka::Headers *headers = RdKafka::Headers::create();
+      headers->add("my header", "header value");
+      headers->add("other header", "yes");
+
       /*
        * Produce message
        */
       RdKafka::ErrorCode resp =
-	producer->produce(topic, partition,
-			  RdKafka::Producer::RK_MSG_COPY /* Copy payload */,
-			  const_cast<char *>(line.c_str()), line.size(),
-			  NULL, NULL);
-      if (resp != RdKafka::ERR_NO_ERROR)
-	std::cerr << "% Produce failed: " <<
-	  RdKafka::err2str(resp) << std::endl;
-      else
-	std::cerr << "% Produced message (" << line.size() << " bytes)" <<
-	  std::endl;
+        producer->produce(topic_str, partition,
+                          RdKafka::Producer::RK_MSG_COPY /* Copy payload */,
+                          /* Value */
+                          const_cast<char *>(line.c_str()), line.size(),
+                          /* Key */
+                          NULL, 0,
+                          /* Timestamp (defaults to now) */
+                          0,
+                          /* Message headers, if any */
+                          headers,
+                          /* Per-message opaque value passed to
+                           * delivery report */
+                          NULL);
+      if (resp != RdKafka::ERR_NO_ERROR) {
+        std::cerr << "% Produce failed: " <<
+          RdKafka::err2str(resp) << std::endl;
+        delete headers; /* Headers are automatically deleted on produce()
+                         * success. */
+      } else {
+        std::cerr << "% Produced message (" << line.size() << " bytes)" <<
+          std::endl;
+      }
 
       producer->poll(0);
     }
     run = true;
 
-    while (run and producer->outq_len() > 0) {
+    while (run && producer->outq_len() > 0) {
       std::cerr << "Waiting for " << producer->outq_len() << std::endl;
       producer->poll(1000);
     }
 
-    delete topic;
     delete producer;
 
 
-  } else {
+  } else if (mode == "C") {
     /*
      * Consumer mode
      */
 
+    conf->set("enable.partition.eof", "true", errstr);
+
+    if(topic_str.empty())
+      goto usage;
+
     /*
      * Create consumer using accumulated global configuration.
      */
@@ -426,42 +602,20 @@ int main (int argc, char **argv) {
       exit(1);
     }
 
+    ExampleConsumeCb ex_consume_cb;
+
     /*
      * Consume messages
      */
     while (run) {
-      RdKafka::Message *msg = consumer->consume(topic, partition, 1000);
-
-      switch (msg->err())
-      {
-        case RdKafka::ERR__TIMED_OUT:
-          break;
-
-        case RdKafka::ERR_NO_ERROR:
-	  /* Real message */
-	  std::cerr << "Read msg at offset " << msg->offset() << std::endl;
-	  if (msg->key()) {
-		  std::cerr << "Key: " << *msg->key() << std::endl;
-	  }
-          printf("%.*s\n",
-                 static_cast<int>(msg->len()),
-                 static_cast<const char *>(msg->payload()));
-	  break;
-
-	case RdKafka::ERR__PARTITION_EOF:
-	  /* Last message */
-	  if (exit_eof)
-	    run = false;
-	  break;
-
-	default:
-	  /* Errors */
-	  std::cerr << "Consume failed: " << msg->errstr() << std::endl;
-	  run = false;
-	}
-
-      delete msg;
-
+      if (use_ccb) {
+        consumer->consume_callback(topic, partition, 1000,
+                                   &ex_consume_cb, &use_ccb);
+      } else {
+        RdKafka::Message *msg = consumer->consume(topic, partition, 1000);
+        msg_consume(msg, NULL);
+        delete msg;
+      }
       consumer->poll(0);
     }
 
@@ -474,8 +628,55 @@ int main (int argc, char **argv) {
 
     delete topic;
     delete consumer;
+  } else {
+    /* Metadata mode */
+
+    /*
+     * Create producer using accumulated global configuration.
+     */
+    RdKafka::Producer *producer = RdKafka::Producer::create(conf, errstr);
+    if (!producer) {
+      std::cerr << "Failed to create producer: " << errstr << std::endl;
+      exit(1);
+    }
+
+    std::cout << "% Created producer " << producer->name() << std::endl;
+
+    /*
+     * Create topic handle.
+     */
+    RdKafka::Topic *topic = NULL;
+    if(!topic_str.empty()) {
+      topic = RdKafka::Topic::create(producer, topic_str, tconf, errstr);
+      if (!topic) {
+        std::cerr << "Failed to create topic: " << errstr << std::endl;
+        exit(1);
+      }
+    }
+
+    while (run) {
+      class RdKafka::Metadata *metadata;
+
+      /* Fetch metadata */
+      RdKafka::ErrorCode err = producer->metadata(!topic, topic,
+                              &metadata, 5000);
+      if (err != RdKafka::ERR_NO_ERROR) {
+        std::cerr << "%% Failed to acquire metadata: "
+                  << RdKafka::err2str(err) << std::endl;
+              run = 0;
+              break;
+      }
+
+      metadata_print(topic_str, metadata);
+
+      delete metadata;
+      run = 0;
+    }
+
   }
 
+  delete conf;
+  delete tconf;
 
   /*
    * Wait for RdKafka to decommission.
