@@ -12,6 +12,7 @@ import tempfile
 import shutil
 import subprocess
 import urllib
+from fnmatch import fnmatch
 from string import Template
 from collections import defaultdict
 import boto3
@@ -19,7 +20,7 @@ from zfile import zfile
 
 
 # Rename token values
-rename_vals = {'plat': {'windows': 'win7'},
+rename_vals = {'plat': {'windows': 'win'},
                'arch': {'x86_64': 'x64',
                         'i386': 'x86',
                         'win32': 'x86'}}
@@ -135,7 +136,7 @@ class Artifacts (object):
 
         # The folder contains the tokens needed to perform
         # matching of project, gitref, etc.
-        rinfo = re.findall(r'(?P<tag>[^-]+)-(?P<val>.*?)__', folder)
+        rinfo = re.findall(r'(?P<tag>[^-]+)-(?P<val>.*?)(?:__|$)', folder)
         if rinfo is None or len(rinfo) == 0:
             print('Incorrect folder/file name format for %s' % folder)
             return None
@@ -175,8 +176,27 @@ class Artifacts (object):
         self.s3 = boto3.resource('s3')
         self.s3_bucket = self.s3.Bucket(s3_bucket)
         self.s3_client = boto3.client('s3')
-        for item in self.s3_client.list_objects(Bucket=s3_bucket, Prefix='librdkafka/').get('Contents'):
-            self.collect_single(item.get('Key'))
+
+        # note: list_objects will return at most 1000 objects per call,
+        #       use continuation token to read full list.
+        cont_token = None
+        more = True
+        while more:
+            if cont_token is not None:
+                res = self.s3_client.list_objects_v2(Bucket=s3_bucket,
+                                                     Prefix='librdkafka/',
+                                                     ContinuationToken=cont_token)
+            else:
+                res = self.s3_client.list_objects_v2(Bucket=s3_bucket,
+                                                     Prefix='librdkafka/')
+
+            if res.get('IsTruncated') == True:
+                cont_token = res.get('NextContinuationToken')
+            else:
+                more = False
+
+            for item in res.get('Contents'):
+                self.collect_single(item.get('Key'))
 
         for a in self.artifacts:
             a.download()
@@ -299,40 +319,53 @@ class NugetPackage (Package):
                 a.info['toolset'] = 'v120'
 
         mappings = [
-            [{'arch': 'x64', 'plat': 'linux', 'fname_startswith': 'librdkafka.tar.gz'}, './include/librdkafka/rdkafka.h', 'build/native/include/librdkafka/rdkafka.h'],
-            [{'arch': 'x64', 'plat': 'linux', 'fname_startswith': 'librdkafka.tar.gz'}, './include/librdkafka/rdkafkacpp.h', 'build/native/include/librdkafka/rdkafkacpp.h'],
+            [{'arch': 'x64', 'plat': 'linux', 'fname_glob': 'librdkafka-gcc.tar.gz'}, './include/librdkafka/rdkafka.h', 'build/native/include/librdkafka/rdkafka.h'],
+            [{'arch': 'x64', 'plat': 'linux', 'fname_glob': 'librdkafka-gcc.tar.gz'}, './include/librdkafka/rdkafkacpp.h', 'build/native/include/librdkafka/rdkafkacpp.h'],
 
-            [{'arch': 'x64', 'plat': 'osx', 'fname_startswith': 'librdkafka.tar.gz'}, './lib/librdkafka.dylib', 'runtimes/osx-x64/native/librdkafka.dylib'],
-            [{'arch': 'x64', 'plat': 'linux', 'fname_startswith': 'librdkafka-debian9.tgz'}, './lib/librdkafka.so.1', 'runtimes/linux-x64/native/debian9-librdkafka.so'],
-            [{'arch': 'x64', 'plat': 'linux', 'fname_startswith': 'librdkafka.tar.gz'}, './lib/librdkafka.so.1', 'runtimes/linux-x64/native/librdkafka.so'],
+            # Travis OSX build
+            [{'arch': 'x64', 'plat': 'osx', 'fname_glob': 'librdkafka-clang.tar.gz'}, './lib/librdkafka.dylib', 'runtimes/osx-x64/native/librdkafka.dylib'],
+            # Travis Debian 9 / Ubuntu 16.04 build
+            [{'arch': 'x64', 'plat': 'linux', 'fname_glob': 'librdkafka-debian9.tgz'}, './lib/librdkafka.so.1', 'runtimes/linux-x64/native/debian9-librdkafka.so'],
+            # Travis Ubuntu 14.04 build
+            [{'arch': 'x64', 'plat': 'linux', 'fname_glob': 'librdkafka-gcc.tar.gz'}, './lib/librdkafka.so.1', 'runtimes/linux-x64/native/librdkafka.so'],
+            # Travis CentOS 7 RPM build
+            [{'arch': 'x64', 'plat': 'linux', 'fname_glob': 'librdkafka1*.x86_64.rpm'}, './usr/lib64/librdkafka.so.1', 'runtimes/linux-x64/native/centos7-librdkafka.so'],
+            # Alpine build
+            [{'arch': 'x64', 'plat': 'linux', 'fname_glob': 'alpine-librdkafka.tgz'}, 'librdkafka.so.1', 'runtimes/linux-x64/native/alpine-librdkafka.so'],
 
-            [{'arch': 'x64', 'plat': 'win7', 'fname_startswith': 'msvcr120.zip'}, 'msvcr120.dll', 'runtimes/win7-x64/native/msvcr120.dll'],
+            # Common Win runtime
+            [{'arch': 'x64', 'plat': 'win', 'fname_glob': 'msvcr120.zip'}, 'msvcr120.dll', 'runtimes/win-x64/native/msvcr120.dll'],
+            [{'arch': 'x64', 'plat': 'win', 'fname_glob': 'msvcr120.zip'}, 'msvcp120.dll', 'runtimes/win-x64/native/msvcp120.dll'],
             # matches librdkafka.redist.{VER}.nupkg
-            [{'arch': 'x64', 'plat': 'win7', 'fname_startswith': 'librdkafka.redist'}, 'build/native/bin/v120/x64/Release/librdkafka.dll', 'runtimes/win7-x64/native/librdkafka.dll'],
-            [{'arch': 'x64', 'plat': 'win7', 'fname_startswith': 'librdkafka.redist'}, 'build/native/bin/v120/x64/Release/librdkafkacpp.dll', 'runtimes/win7-x64/native/librdkafkacpp.dll'],
-            [{'arch': 'x64', 'plat': 'win7', 'fname_startswith': 'librdkafka.redist'}, 'build/native/bin/v120/x64/Release/zlib.dll', 'runtimes/win7-x64/native/zlib.dll'],
+            [{'arch': 'x64', 'plat': 'win', 'fname_glob': 'librdkafka.redist*'}, 'build/native/bin/v120/x64/Release/librdkafka.dll', 'runtimes/win-x64/native/librdkafka.dll'],
+            [{'arch': 'x64', 'plat': 'win', 'fname_glob': 'librdkafka.redist*'}, 'build/native/bin/v120/x64/Release/librdkafkacpp.dll', 'runtimes/win-x64/native/librdkafkacpp.dll'],
+            [{'arch': 'x64', 'plat': 'win', 'fname_glob': 'librdkafka.redist*'}, 'build/native/bin/v120/x64/Release/zlib.dll', 'runtimes/win-x64/native/zlib.dll'],
+            [{'arch': 'x64', 'plat': 'win', 'fname_glob': 'librdkafka.redist*'}, 'build/native/bin/v120/x64/Release/libzstd.dll', 'runtimes/win-x64/native/libzstd.dll'],
             # matches librdkafka.{VER}.nupkg
-            [{'arch': 'x64', 'plat': 'win7', 'fname_startswith': 'librdkafka', 'fname_excludes': ['redist', 'symbols']},
-             'build/native/lib/v120/x64/Release/librdkafka.lib', 'build/native/lib/win7/x64/win7-x64-Release/v120/librdkafka.lib'],
-            [{'arch': 'x64', 'plat': 'win7', 'fname_startswith': 'librdkafka', 'fname_excludes': ['redist', 'symbols']},
-             'build/native/lib/v120/x64/Release/librdkafkacpp.lib', 'build/native/lib/win7/x64/win7-x64-Release/v120/librdkafkacpp.lib'],
+            [{'arch': 'x64', 'plat': 'win', 'fname_glob': 'librdkafka*', 'fname_excludes': ['redist', 'symbols']},
+             'build/native/lib/v120/x64/Release/librdkafka.lib', 'build/native/lib/win/x64/win-x64-Release/v120/librdkafka.lib'],
+            [{'arch': 'x64', 'plat': 'win', 'fname_glob': 'librdkafka*', 'fname_excludes': ['redist', 'symbols']},
+             'build/native/lib/v120/x64/Release/librdkafkacpp.lib', 'build/native/lib/win/x64/win-x64-Release/v120/librdkafkacpp.lib'],
 
-            [{'arch': 'x86', 'plat': 'win7', 'fname_startswith': 'msvcr120.zip'}, 'msvcr120.dll', 'runtimes/win7-x86/native/msvcr120.dll'],
+            [{'arch': 'x86', 'plat': 'win', 'fname_glob': 'msvcr120.zip'}, 'msvcr120.dll', 'runtimes/win-x86/native/msvcr120.dll'],
+            [{'arch': 'x86', 'plat': 'win', 'fname_glob': 'msvcr120.zip'}, 'msvcp120.dll', 'runtimes/win-x86/native/msvcp120.dll'],
             # matches librdkafka.redist.{VER}.nupkg
-            [{'arch': 'x86', 'plat': 'win7', 'fname_startswith': 'librdkafka.redist'}, 'build/native/bin/v120/Win32/Release/librdkafka.dll', 'runtimes/win7-x86/native/librdkafka.dll'],
-            [{'arch': 'x86', 'plat': 'win7', 'fname_startswith': 'librdkafka.redist'}, 'build/native/bin/v120/Win32/Release/librdkafkacpp.dll', 'runtimes/win7-x86/native/librdkafkacpp.dll'],
-            [{'arch': 'x86', 'plat': 'win7', 'fname_startswith': 'librdkafka.redist'}, 'build/native/bin/v120/Win32/Release/zlib.dll', 'runtimes/win7-x86/native/zlib.dll'],
+            [{'arch': 'x86', 'plat': 'win', 'fname_glob': 'librdkafka.redist*'}, 'build/native/bin/v120/Win32/Release/librdkafka.dll', 'runtimes/win-x86/native/librdkafka.dll'],
+            [{'arch': 'x86', 'plat': 'win', 'fname_glob': 'librdkafka.redist*'}, 'build/native/bin/v120/Win32/Release/librdkafkacpp.dll', 'runtimes/win-x86/native/librdkafkacpp.dll'],
+            [{'arch': 'x86', 'plat': 'win', 'fname_glob': 'librdkafka.redist*'}, 'build/native/bin/v120/Win32/Release/zlib.dll', 'runtimes/win-x86/native/zlib.dll'],
+            [{'arch': 'x86', 'plat': 'win', 'fname_glob': 'librdkafka.redist*'}, 'build/native/bin/v120/Win32/Release/libzstd.dll', 'runtimes/win-x86/native/libzstd.dll'],
+
             # matches librdkafka.{VER}.nupkg
-            [{'arch': 'x86', 'plat': 'win7', 'fname_startswith': 'librdkafka', 'fname_excludes': ['redist', 'symbols']}, 
-            'build/native/lib/v120/Win32/Release/librdkafka.lib', 'build/native/lib/win7/x86/win7-x86-Release/v120/librdkafka.lib'],
-            [{'arch': 'x86', 'plat': 'win7', 'fname_startswith': 'librdkafka', 'fname_excludes': ['redist', 'symbols']}, 
-            'build/native/lib/v120/Win32/Release/librdkafkacpp.lib', 'build/native/lib/win7/x86/win7-x86-Release/v120/librdkafkacpp.lib']
+            [{'arch': 'x86', 'plat': 'win', 'fname_glob': 'librdkafka*', 'fname_excludes': ['redist', 'symbols']},
+            'build/native/lib/v120/Win32/Release/librdkafka.lib', 'build/native/lib/win/x86/win-x86-Release/v120/librdkafka.lib'],
+            [{'arch': 'x86', 'plat': 'win', 'fname_glob': 'librdkafka*', 'fname_excludes': ['redist', 'symbols']},
+            'build/native/lib/v120/Win32/Release/librdkafkacpp.lib', 'build/native/lib/win/x86/win-x86-Release/v120/librdkafkacpp.lib']
         ]
 
         for m in mappings:
             attributes = m[0]
-            fname_startswith = attributes['fname_startswith']
-            del attributes['fname_startswith']
+            fname_glob = attributes['fname_glob']
+            del attributes['fname_glob']
             fname_excludes = []
             if 'fname_excludes' in attributes:
                 fname_excludes = attributes['fname_excludes']
@@ -347,7 +380,7 @@ class NugetPackage (Package):
                         found = False
                         break
 
-                if not a.fname.startswith(fname_startswith):
+                if not fnmatch(a.fname, fname_glob):
                     found = False
 
                 for exclude in fname_excludes:
@@ -360,7 +393,7 @@ class NugetPackage (Package):
                     break
 
             if artifact is None:
-                raise Exception('unable to find file in archive %s with tags %s that starts with "%s"' % (a.fname, str(attributes), fname_startswith))
+                raise Exception('unable to find artifact with tags %s matching "%s"' % (str(attributes), fname_glob))
 
             outf = os.path.join(self.stpath, m[2])
             member = m[1]
@@ -377,7 +410,7 @@ class NugetPackage (Package):
         subprocess.check_call("./nuget.sh pack %s -BasePath '%s' -NonInteractive" %  \
                               (os.path.join(self.stpath, 'librdkafka.redist.nuspec'),
                                self.stpath), shell=True)
-                               
+
         return 'librdkafka.redist.%s.nupkg' % vless_version
 
     def verify (self, path):
@@ -389,33 +422,39 @@ class NugetPackage (Package):
             "build/native/librdkafka.redist.targets",
             "build/native/include/librdkafka/rdkafka.h",
             "build/native/include/librdkafka/rdkafkacpp.h",
-            "build/native/lib/win7/x64/win7-x64-Release/v120/librdkafka.lib",
-            "build/native/lib/win7/x64/win7-x64-Release/v120/librdkafkacpp.lib",
-            "build/native/lib/win7/x86/win7-x86-Release/v120/librdkafka.lib",
-            "build/native/lib/win7/x86/win7-x86-Release/v120/librdkafkacpp.lib",
+            "build/native/lib/win/x64/win-x64-Release/v120/librdkafka.lib",
+            "build/native/lib/win/x64/win-x64-Release/v120/librdkafkacpp.lib",
+            "build/native/lib/win/x86/win-x86-Release/v120/librdkafka.lib",
+            "build/native/lib/win/x86/win-x86-Release/v120/librdkafkacpp.lib",
+            "runtimes/linux-x64/native/centos7-librdkafka.so",
             "runtimes/linux-x64/native/debian9-librdkafka.so",
+            "runtimes/linux-x64/native/alpine-librdkafka.so",
             "runtimes/linux-x64/native/librdkafka.so",
             "runtimes/osx-x64/native/librdkafka.dylib",
-            "runtimes/win7-x64/native/librdkafka.dll",
-            "runtimes/win7-x64/native/librdkafkacpp.dll",
-            "runtimes/win7-x64/native/msvcr120.dll",
-            "runtimes/win7-x64/native/zlib.dll",
-            "runtimes/win7-x86/native/librdkafka.dll",
-            "runtimes/win7-x86/native/librdkafkacpp.dll",
-            "runtimes/win7-x86/native/msvcr120.dll",
-            "runtimes/win7-x86/native/zlib.dll"]
+            "runtimes/win-x64/native/librdkafka.dll",
+            "runtimes/win-x64/native/librdkafkacpp.dll",
+            "runtimes/win-x64/native/msvcr120.dll",
+            "runtimes/win-x64/native/msvcp120.dll",
+            "runtimes/win-x64/native/zlib.dll",
+            "runtimes/win-x64/native/libzstd.dll",
+            "runtimes/win-x86/native/librdkafka.dll",
+            "runtimes/win-x86/native/librdkafkacpp.dll",
+            "runtimes/win-x86/native/msvcr120.dll",
+            "runtimes/win-x86/native/msvcp120.dll",
+            "runtimes/win-x86/native/zlib.dll",
+            "runtimes/win-x86/native/libzstd.dll"]
 
-        missing = list()		
-        with zfile.ZFile(path, 'r') as zf:		
-            print('Verifying %s:' % path)		
-        
-            # Zipfiles may url-encode filenames, unquote them before matching.		
-            pkgd = [urllib.unquote(x) for x in zf.getnames()]		
-            missing = [x for x in expect if x not in pkgd]		
-        
-        if len(missing) > 0:		
-            print('Missing files in package %s:\n%s' % (path, '\n'.join(missing)))		
-            return False		
-        else:		
-            print('OK - %d expected files found' % len(expect))		
+        missing = list()
+        with zfile.ZFile(path, 'r') as zf:
+            print('Verifying %s:' % path)
+
+            # Zipfiles may url-encode filenames, unquote them before matching.
+            pkgd = [urllib.unquote(x) for x in zf.getnames()]
+            missing = [x for x in expect if x not in pkgd]
+
+        if len(missing) > 0:
+            print('Missing files in package %s:\n%s' % (path, '\n'.join(missing)))
+            return False
+        else:
+            print('OK - %d expected files found' % len(expect))
             return True
