@@ -68,13 +68,22 @@ void RdKafka::log_cb_trampoline (const rd_kafka_t *rk, int level,
 void RdKafka::error_cb_trampoline (rd_kafka_t *rk, int err,
                                    const char *reason, void *opaque) {
   RdKafka::HandleImpl *handle = static_cast<RdKafka::HandleImpl *>(opaque);
+  char errstr[512];
+  bool is_fatal = false;
 
+  if (err == RD_KAFKA_RESP_ERR__FATAL) {
+    /* Translate to underlying fatal error code and string */
+    err = rd_kafka_fatal_error(rk, errstr, sizeof(errstr));
+    if (err)
+      reason = errstr;
+    is_fatal = true;
+  }
   RdKafka::EventImpl event(RdKafka::Event::EVENT_ERROR,
                            static_cast<RdKafka::ErrorCode>(err),
                            RdKafka::Event::EVENT_SEVERITY_ERROR,
                            NULL,
                            reason);
-
+  event.fatal_ = is_fatal;
   handle->event_cb_->event_cb(event);
 }
 
@@ -123,6 +132,50 @@ int RdKafka::open_cb_trampoline (const char *pathname, int flags, mode_t mode,
   return handle->open_cb_->open_cb(pathname, flags, static_cast<int>(mode));
 }
 
+void
+RdKafka::oauthbearer_token_refresh_cb_trampoline (rd_kafka_t *rk,
+                                                  const char *oauthbearer_config,
+                                                  void *opaque) {
+  RdKafka::HandleImpl *handle = static_cast<RdKafka::HandleImpl *>(opaque);
+
+  handle->oauthbearer_token_refresh_cb_->
+    oauthbearer_token_refresh_cb(std::string(oauthbearer_config ?
+                                             oauthbearer_config : ""));
+}
+
+
+int RdKafka::ssl_cert_verify_cb_trampoline (rd_kafka_t *rk,
+                                            const char *broker_name,
+                                            int32_t broker_id,
+                                            int *x509_error,
+                                            int depth,
+                                            const char *buf, size_t size,
+                                            char *errstr, size_t errstr_size,
+                                            void *opaque) {
+  RdKafka::HandleImpl *handle = static_cast<RdKafka::HandleImpl *>(opaque);
+  std::string errbuf;
+
+  bool res = 0 != handle->ssl_cert_verify_cb_->
+    ssl_cert_verify_cb(std::string(broker_name), broker_id,
+                       x509_error,
+                       depth,
+                       buf, size,
+                       errbuf);
+
+  if (res)
+    return (int)res;
+
+  size_t errlen = errbuf.size() > errstr_size - 1 ?
+    errstr_size - 1 : errbuf.size();
+
+  memcpy(errstr, errbuf.c_str(), errlen);
+  if (errstr_size > 0)
+    errstr[errlen] = '\0';
+
+  return (int)res;
+}
+
+
 RdKafka::ErrorCode RdKafka::HandleImpl::metadata (bool all_topics,
                                                   const Topic *only_rkt,
                                                   Metadata **metadatap, 
@@ -180,22 +233,31 @@ RdKafka::rebalance_cb_trampoline (rd_kafka_t *rk,
 
 
 void
-RdKafka::offset_commit_cb_trampoline (
+RdKafka::offset_commit_cb_trampoline0 (
     rd_kafka_t *rk,
     rd_kafka_resp_err_t err,
     rd_kafka_topic_partition_list_t *c_offsets, void *opaque) {
-  RdKafka::HandleImpl *handle = static_cast<RdKafka::HandleImpl *>(opaque);
+  OffsetCommitCb *cb = static_cast<RdKafka::OffsetCommitCb *>(opaque);
   std::vector<RdKafka::TopicPartition*> offsets;
 
   if (c_offsets)
     c_parts_to_partitions(c_offsets, offsets);
 
-  handle->offset_commit_cb_->
-      offset_commit_cb(static_cast<RdKafka::ErrorCode>(err), offsets);
+  cb->offset_commit_cb(static_cast<RdKafka::ErrorCode>(err), offsets);
 
   free_partition_vector(offsets);
 }
 
+static void
+offset_commit_cb_trampoline (
+    rd_kafka_t *rk,
+    rd_kafka_resp_err_t err,
+    rd_kafka_topic_partition_list_t *c_offsets, void *opaque) {
+  RdKafka::HandleImpl *handle = static_cast<RdKafka::HandleImpl *>(opaque);
+  RdKafka::offset_commit_cb_trampoline0(rk, err, c_offsets,
+                                        handle->offset_commit_cb_);
+}
+
 
 void RdKafka::HandleImpl::set_common_config (RdKafka::ConfImpl *confimpl) {
 
@@ -213,12 +275,25 @@ void RdKafka::HandleImpl::set_common_config (RdKafka::ConfImpl *confimpl) {
     event_cb_ = confimpl->event_cb_;
   }
 
+  if (confimpl->oauthbearer_token_refresh_cb_) {
+    rd_kafka_conf_set_oauthbearer_token_refresh_cb(
+          confimpl->rk_conf_,
+          RdKafka::oauthbearer_token_refresh_cb_trampoline);
+      oauthbearer_token_refresh_cb_ = confimpl->oauthbearer_token_refresh_cb_;
+  }
+
   if (confimpl->socket_cb_) {
     rd_kafka_conf_set_socket_cb(confimpl->rk_conf_,
                                 RdKafka::socket_cb_trampoline);
     socket_cb_ = confimpl->socket_cb_;
   }
 
+  if (confimpl->ssl_cert_verify_cb_) {
+    rd_kafka_conf_set_ssl_cert_verify_cb(confimpl->rk_conf_,
+                                         RdKafka::ssl_cert_verify_cb_trampoline);
+      ssl_cert_verify_cb_ = confimpl->ssl_cert_verify_cb_;
+  }
+
   if (confimpl->open_cb_) {
 #ifndef _MSC_VER
     rd_kafka_conf_set_open_cb(confimpl->rk_conf_, RdKafka::open_cb_trampoline);
@@ -234,7 +309,7 @@ void RdKafka::HandleImpl::set_common_config (RdKafka::ConfImpl *confimpl) {
 
   if (confimpl->offset_commit_cb_) {
     rd_kafka_conf_set_offset_commit_cb(confimpl->rk_conf_,
-                                   RdKafka::offset_commit_cb_trampoline);
+                                       offset_commit_cb_trampoline);
     offset_commit_cb_ = confimpl->offset_commit_cb_;
   }
 
@@ -282,6 +357,32 @@ RdKafka::HandleImpl::resume (std::vector<RdKafka::TopicPartition*> &partitions)
   return static_cast<RdKafka::ErrorCode>(err);
 }
 
+RdKafka::Queue *
+RdKafka::HandleImpl::get_partition_queue (const TopicPartition *part) {
+  rd_kafka_queue_t *rkqu;
+  rkqu = rd_kafka_queue_get_partition(rk_,
+                                      part->topic().c_str(),
+                                      part->partition());
+
+  if (rkqu == NULL)
+    return NULL;
+
+  RdKafka::QueueImpl *queueimpl = new RdKafka::QueueImpl;
+  queueimpl->queue_ = rkqu;
+
+  return queueimpl;
+}
+
+RdKafka::ErrorCode
+RdKafka::HandleImpl::set_log_queue (RdKafka::Queue *queue) {
+        rd_kafka_queue_t *rkqu = NULL;
+        if (queue) {
+                QueueImpl *queueimpl = dynamic_cast<QueueImpl *>(queue);
+                rkqu = queueimpl->queue_;
+        }
+        return static_cast<RdKafka::ErrorCode>(
+                rd_kafka_set_log_queue(rk_, rkqu));
+}
 
 namespace RdKafka {
 
@@ -289,7 +390,7 @@ rd_kafka_topic_partition_list_t *
 partitions_to_c_parts (const std::vector<RdKafka::TopicPartition*> &partitions){
   rd_kafka_topic_partition_list_t *c_parts;
 
-  c_parts = rd_kafka_topic_partition_list_new(partitions.size());
+  c_parts = rd_kafka_topic_partition_list_new((int)partitions.size());
 
   for (unsigned int i = 0 ; i < partitions.size() ; i++) {
     const RdKafka::TopicPartitionImpl *tpi =
