@@ -40,6 +40,14 @@ extern "C" {
 #include "../src/rdkafka.h"
 }
 
+#ifdef _MSC_VER
+/* Visual Studio */
+#include "../src/win32_config.h"
+#else
+/* POSIX / UNIX based systems */
+#include "../config.h" /* mklove output */
+#endif
+
 #ifdef _MSC_VER
 typedef int mode_t;
 #pragma warning(disable : 4250)
@@ -48,7 +56,6 @@ typedef int mode_t;
 
 namespace RdKafka {
 
-
 void consume_cb_trampoline(rd_kafka_message_t *msg, void *opaque);
 void log_cb_trampoline (const rd_kafka_t *rk, int level,
                         const char *fac, const char *buf);
@@ -70,6 +77,19 @@ void offset_commit_cb_trampoline0 (
         rd_kafka_t *rk,
         rd_kafka_resp_err_t err,
         rd_kafka_topic_partition_list_t *c_offsets, void *opaque);
+void oauthbearer_token_refresh_cb_trampoline (rd_kafka_t *rk,
+                                              const char *oauthbearer_config,
+                                              void *opaque);
+
+ int ssl_cert_verify_cb_trampoline (
+         rd_kafka_t *rk,
+         const char *broker_name,
+         int32_t broker_id,
+         int *x509_error,
+         int depth,
+         const char *buf, size_t size,
+         char *errstr, size_t errstr_size,
+         void *opaque);
 
 rd_kafka_topic_partition_list_t *
     partitions_to_c_parts (const std::vector<TopicPartition*> &partitions);
@@ -108,6 +128,8 @@ class EventImpl : public Event {
   int         broker_id () const { return id_; }
   int         throttle_time () const { return throttle_time_; }
 
+  bool        fatal () const { return fatal_; }
+
   Type        type_;
   ErrorCode   err_;
   Severity    severity_;
@@ -115,27 +137,162 @@ class EventImpl : public Event {
   std::string str_;         /* reused for THROTTLE broker_name */
   int         id_;
   int         throttle_time_;
+  bool        fatal_;
 };
 
 
+class HeadersImpl : public Headers {
+ public:
+  HeadersImpl ():
+  headers_ (rd_kafka_headers_new(8)) {}
+
+  HeadersImpl (rd_kafka_headers_t *headers):
+  headers_ (headers) {}
+
+  HeadersImpl (const std::vector<Header> &headers) {
+    if (headers.size() > 0) {
+      headers_ = rd_kafka_headers_new(headers.size());
+      from_vector(headers);
+    } else {
+      headers_ = rd_kafka_headers_new(8);
+    }
+  }
+
+  ~HeadersImpl() {
+    if (headers_) {
+      rd_kafka_headers_destroy(headers_);
+    }
+  }
+
+  ErrorCode add(const std::string& key, const char *value) {
+    rd_kafka_resp_err_t err;
+    err = rd_kafka_header_add(headers_,
+                              key.c_str(), key.size(),
+                              value, -1);
+    return static_cast<RdKafka::ErrorCode>(err);
+  }
+
+  ErrorCode add(const std::string& key, const void *value, size_t value_size) {
+    rd_kafka_resp_err_t err;
+    err = rd_kafka_header_add(headers_,
+                              key.c_str(), key.size(),
+                              value, value_size);
+    return static_cast<RdKafka::ErrorCode>(err);
+  }
+
+  ErrorCode add(const std::string &key, const std::string &value) {
+    rd_kafka_resp_err_t err;
+    err = rd_kafka_header_add(headers_,
+                              key.c_str(), key.size(),
+                              value.c_str(), value.size());
+    return static_cast<RdKafka::ErrorCode>(err);
+  }
+
+  ErrorCode add(const Header &header) {
+    rd_kafka_resp_err_t err;
+    err = rd_kafka_header_add(headers_,
+                              header.key().c_str(), header.key().size(),
+                              header.value(), header.value_size());
+    return static_cast<RdKafka::ErrorCode>(err);
+  }
+
+  ErrorCode remove(const std::string& key) {
+    rd_kafka_resp_err_t err;
+    err = rd_kafka_header_remove (headers_, key.c_str());
+    return static_cast<RdKafka::ErrorCode>(err);
+  }
+
+  std::vector<Headers::Header> get(const std::string &key) const {
+    std::vector<Headers::Header> headers;
+    const void *value;
+    size_t size;
+    rd_kafka_resp_err_t err;
+    for (size_t idx = 0;
+         !(err = rd_kafka_header_get(headers_, idx, key.c_str(),
+                                     &value, &size)) ;
+         idx++) {
+      headers.push_back(Headers::Header(key, value, size));
+    }
+    return headers;
+  }
+
+  Headers::Header get_last(const std::string& key) const {
+    const void *value;
+    size_t size;
+    rd_kafka_resp_err_t err;
+    err = rd_kafka_header_get_last(headers_, key.c_str(), &value, &size);
+    return Headers::Header(key, value, size,
+                           static_cast<RdKafka::ErrorCode>(err));
+  }
+
+  std::vector<Headers::Header> get_all() const {
+    std::vector<Headers::Header> headers;
+    size_t idx = 0;
+    const char *name;
+    const void *valuep;
+    size_t size;
+    while (!rd_kafka_header_get_all(headers_, idx++,
+                                    &name, &valuep, &size)) {
+      headers.push_back(Headers::Header(name, valuep, size));
+    }
+    return headers;
+  }
+
+  size_t size() const {
+    return rd_kafka_header_cnt(headers_);
+  }
+
+  /** @brief Reset the C headers pointer to NULL. */
+  void c_headers_destroyed() {
+    headers_ = NULL;
+  }
+
+  /** @returns the underlying C headers, or NULL. */
+  rd_kafka_headers_t *c_ptr() {
+    return headers_;
+  }
+
+
+private:
+  void from_vector(const std::vector<Header> &headers) {
+    if (headers.size() == 0)
+      return;
+    for (std::vector<Header>::const_iterator it = headers.begin();
+         it != headers.end(); it++)
+      this->add(*it);
+  }
+
+  HeadersImpl(HeadersImpl const&) /*= delete*/;
+  HeadersImpl& operator=(HeadersImpl const&) /*= delete*/;
+
+  rd_kafka_headers_t *headers_;
+};
+
+
+
 class MessageImpl : public Message {
  public:
   ~MessageImpl () {
     if (free_rkmessage_)
       rd_kafka_message_destroy(const_cast<rd_kafka_message_t *>(rkmessage_));
     if (key_)
-            delete key_;
+      delete key_;
+    if (headers_)
+      delete headers_;
   };
 
   MessageImpl (RdKafka::Topic *topic, rd_kafka_message_t *rkmessage):
-  topic_(topic), rkmessage_(rkmessage), free_rkmessage_(true), key_(NULL) {}
+  topic_(topic), rkmessage_(rkmessage), free_rkmessage_(true), key_(NULL),
+  headers_(NULL) {}
 
   MessageImpl (RdKafka::Topic *topic, rd_kafka_message_t *rkmessage,
                bool dofree):
-  topic_(topic), rkmessage_(rkmessage), free_rkmessage_(dofree), key_(NULL) { }
+  topic_(topic), rkmessage_(rkmessage), free_rkmessage_(dofree), key_(NULL),
+  headers_(NULL) {}
 
   MessageImpl (rd_kafka_message_t *rkmessage):
-  topic_(NULL), rkmessage_(rkmessage), free_rkmessage_(true), key_(NULL) {
+  topic_(NULL), rkmessage_(rkmessage), free_rkmessage_(true), key_(NULL),
+  headers_(NULL) {
     if (rkmessage->rkt) {
       /* Possibly NULL */
       topic_ = static_cast<Topic *>(rd_kafka_topic_opaque(rkmessage->rkt));
@@ -144,7 +301,7 @@ class MessageImpl : public Message {
 
   /* Create errored message */
   MessageImpl (RdKafka::Topic *topic, RdKafka::ErrorCode err):
-  topic_(topic), free_rkmessage_(false), key_(NULL) {
+  topic_(topic), free_rkmessage_(false), key_(NULL), headers_(NULL) {
     rkmessage_ = &rkmessage_err_;
     memset(&rkmessage_err_, 0, sizeof(rkmessage_err_));
     rkmessage_err_.err = static_cast<rd_kafka_resp_err_t>(err);
@@ -206,6 +363,33 @@ class MessageImpl : public Message {
           return rkmessage_;
   }
 
+  Status status () const {
+          return static_cast<Status>(rd_kafka_message_status(rkmessage_));
+  }
+
+  Headers *headers () {
+    ErrorCode err;
+    return headers(&err);
+  }
+
+  Headers *headers (ErrorCode *err) {
+    *err = ERR_NO_ERROR;
+
+    if (!headers_) {
+      rd_kafka_headers_t *c_hdrs;
+      rd_kafka_resp_err_t c_err;
+
+      if ((c_err = rd_kafka_message_detach_headers(rkmessage_, &c_hdrs))) {
+        *err = static_cast<RdKafka::ErrorCode>(c_err);
+        return NULL;
+      }
+
+      headers_ = new HeadersImpl(c_hdrs);
+    }
+
+    return headers_;
+  }
+
   RdKafka::Topic *topic_;
   rd_kafka_message_t *rkmessage_;
   bool free_rkmessage_;
@@ -218,6 +402,8 @@ private:
   /* "delete" copy ctor + copy assignment, for safety of key_ */
   MessageImpl(MessageImpl const&) /*= delete*/;
   MessageImpl& operator=(MessageImpl const&) /*= delete*/;
+
+  RdKafka::Headers *headers_;
 };
 
 
@@ -233,6 +419,8 @@ class ConfImpl : public Conf {
       partitioner_kp_cb_(NULL),
       rebalance_cb_(NULL),
       offset_commit_cb_(NULL),
+      oauthbearer_token_refresh_cb_(NULL),
+      ssl_cert_verify_cb_(NULL),
       rk_conf_(NULL),
       rkt_conf_(NULL){}
   ~ConfImpl () {
@@ -262,6 +450,23 @@ class ConfImpl : public Conf {
     return Conf::CONF_OK;
   }
 
+  Conf::ConfResult set (const std::string &name,
+                        OAuthBearerTokenRefreshCb *oauthbearer_token_refresh_cb,
+                        std::string &errstr) {
+    if (name != "oauthbearer_token_refresh_cb") {
+      errstr = "Invalid value type, expected RdKafka::OAuthBearerTokenRefreshCb";
+      return Conf::CONF_INVALID;
+    }
+
+    if (!rk_conf_) {
+      errstr = "Requires RdKafka::Conf::CONF_GLOBAL object";
+      return Conf::CONF_INVALID;
+    }
+
+    oauthbearer_token_refresh_cb_ = oauthbearer_token_refresh_cb;
+    return Conf::CONF_OK;
+  }
+
   Conf::ConfResult set (const std::string &name, EventCb *event_cb,
                         std::string &errstr) {
     if (name != "event_cb") {
@@ -402,6 +607,49 @@ class ConfImpl : public Conf {
     return Conf::CONF_OK;
   }
 
+
+  Conf::ConfResult set (const std::string &name,
+                        SslCertificateVerifyCb *ssl_cert_verify_cb,
+                        std::string &errstr) {
+    if (name != "ssl_cert_verify_cb") {
+      errstr = "Invalid value type, expected RdKafka::SslCertificateVerifyCb";
+      return Conf::CONF_INVALID;
+    }
+
+    if (!rk_conf_) {
+      errstr = "Requires RdKafka::Conf::CONF_GLOBAL object";
+      return Conf::CONF_INVALID;
+    }
+
+    ssl_cert_verify_cb_ = ssl_cert_verify_cb;
+    return Conf::CONF_OK;
+  }
+
+  Conf::ConfResult set_ssl_cert (RdKafka::CertificateType cert_type,
+                                 RdKafka::CertificateEncoding cert_enc,
+                                 const void *buffer, size_t size,
+                                 std::string &errstr) {
+    rd_kafka_conf_res_t res;
+    char errbuf[512];
+
+    if (!rk_conf_) {
+      errstr = "Requires RdKafka::Conf::CONF_GLOBAL object";
+      return Conf::CONF_INVALID;
+    }
+
+    res = rd_kafka_conf_set_ssl_cert(
+        rk_conf_,
+        static_cast<rd_kafka_cert_type_t>(cert_type),
+        static_cast<rd_kafka_cert_enc_t>(cert_enc),
+        buffer, size, errbuf, sizeof(errbuf));
+
+    if (res != RD_KAFKA_CONF_OK)
+      errstr = errbuf;
+
+    return static_cast<Conf::ConfResult>(res);
+  }
+
+
   Conf::ConfResult get(const std::string &name, std::string &value) const {
     if (name.compare("dr_cb") == 0 ||
         name.compare("event_cb") == 0 ||
@@ -410,7 +658,9 @@ class ConfImpl : public Conf {
         name.compare("socket_cb") == 0 ||
         name.compare("open_cb") == 0 ||
         name.compare("rebalance_cb") == 0 ||
-        name.compare("offset_commit_cb") == 0 ) {
+        name.compare("offset_commit_cb") == 0 ||
+        name.compare("oauthbearer_token_refresh_cb") == 0 ||
+        name.compare("ssl_cert_verify_cb") == 0) {
       return Conf::CONF_INVALID;
     }
     rd_kafka_conf_res_t res = RD_KAFKA_CONF_INVALID;
@@ -433,7 +683,7 @@ class ConfImpl : public Conf {
                               tmpValue, &size);
     else if (rkt_conf_)
       res = rd_kafka_topic_conf_get(rkt_conf_,
-                                    name.c_str(), NULL, &size);
+                                    name.c_str(), tmpValue, &size);
 
     if (res == RD_KAFKA_CONF_OK)
       value.assign(tmpValue);
@@ -449,6 +699,14 @@ class ConfImpl : public Conf {
       return Conf::CONF_OK;
   }
 
+  Conf::ConfResult get(
+    OAuthBearerTokenRefreshCb *&oauthbearer_token_refresh_cb) const {
+      if (!rk_conf_)
+          return Conf::CONF_INVALID;
+      oauthbearer_token_refresh_cb = this->oauthbearer_token_refresh_cb_;
+      return Conf::CONF_OK;
+  }
+
   Conf::ConfResult get(EventCb *&event_cb) const {
       if (!rk_conf_)
 	  return Conf::CONF_INVALID;
@@ -498,7 +756,12 @@ class ConfImpl : public Conf {
       return Conf::CONF_OK;
     }
 
-
+  Conf::ConfResult get(SslCertificateVerifyCb *&ssl_cert_verify_cb) const {
+      if (!rk_conf_)
+              return Conf::CONF_INVALID;
+      ssl_cert_verify_cb = this->ssl_cert_verify_cb_;
+      return Conf::CONF_OK;
+  }
 
   std::list<std::string> *dump ();
 
@@ -519,6 +782,19 @@ class ConfImpl : public Conf {
     return Conf::CONF_OK;
   }
 
+  struct rd_kafka_conf_s *c_ptr_global () {
+    if (conf_type_ == CONF_GLOBAL)
+      return rk_conf_;
+    else
+      return NULL;
+  }
+
+  struct rd_kafka_topic_conf_s *c_ptr_topic () {
+    if (conf_type_ == CONF_TOPIC)
+      return rkt_conf_;
+    else
+      return NULL;
+  }
 
   ConsumeCb *consume_cb_;
   DeliveryReportCb *dr_cb_;
@@ -529,6 +805,8 @@ class ConfImpl : public Conf {
   PartitionerKeyPointerCb *partitioner_kp_cb_;
   RebalanceCb *rebalance_cb_;
   OffsetCommitCb *offset_commit_cb_;
+  OAuthBearerTokenRefreshCb *oauthbearer_token_refresh_cb_;
+  SslCertificateVerifyCb *ssl_cert_verify_cb_;
   ConfType conf_type_;
   rd_kafka_conf_t *rk_conf_;
   rd_kafka_topic_conf_t *rkt_conf_;
@@ -611,6 +889,49 @@ class HandleImpl : virtual public Handle {
           return rd_kafka_controllerid(rk_, timeout_ms);
   }
 
+  ErrorCode fatal_error (std::string &errstr) {
+          char errbuf[512];
+          RdKafka::ErrorCode err =
+                  static_cast<RdKafka::ErrorCode>(
+                          rd_kafka_fatal_error(rk_, errbuf, sizeof(errbuf)));
+          if (err)
+                  errstr = errbuf;
+          return err;
+  }
+
+  ErrorCode oauthbearer_set_token (const std::string &token_value,
+                                   int64_t md_lifetime_ms,
+                                   const std::string &md_principal_name,
+                                   const std::list<std::string> &extensions,
+                                   std::string &errstr) {
+          char errbuf[512];
+          ErrorCode err;
+          const char **extensions_copy = new const char *[extensions.size()];
+          int elem = 0;
+
+          for (std::list<std::string>::const_iterator it = extensions.begin();
+              it != extensions.end(); it++)
+                  extensions_copy[elem++] = it->c_str();
+          err = static_cast<ErrorCode>(rd_kafka_oauthbearer_set_token(
+                                               rk_, token_value.c_str(),
+                                               md_lifetime_ms,
+                                               md_principal_name.c_str(),
+                                               extensions_copy,
+                                               extensions.size(),
+                                               errbuf, sizeof(errbuf)));
+          free(extensions_copy);
+
+          if (err != ERR_NO_ERROR)
+              errstr = errbuf;
+
+          return err;
+  }
+
+  ErrorCode oauthbearer_set_token_failure(const std::string &errstr) {
+          return static_cast<ErrorCode>(rd_kafka_oauthbearer_set_token_failure(
+                                                rk_, errstr.c_str()));
+  };
+
 
   rd_kafka_t *rk_;
   /* All Producer and Consumer callbacks must reside in HandleImpl and
@@ -626,6 +947,8 @@ class HandleImpl : virtual public Handle {
   PartitionerKeyPointerCb *partitioner_kp_cb_;
   RebalanceCb *rebalance_cb_;
   OffsetCommitCb *offset_commit_cb_;
+  OAuthBearerTokenRefreshCb *oauthbearer_token_refresh_cb_;
+  SslCertificateVerifyCb *ssl_cert_verify_cb_;
 };
 
 
@@ -851,7 +1174,9 @@ class QueueImpl : virtual public Queue {
 class ConsumerImpl : virtual public Consumer, virtual public HandleImpl {
  public:
   ~ConsumerImpl () {
-    rd_kafka_destroy(rk_); };
+    if (rk_)
+      rd_kafka_destroy(rk_);
+  };
   static Consumer *create (Conf *conf, std::string &errstr);
 
   ErrorCode start (Topic *topic, int32_t partition, int64_t offset);
@@ -892,11 +1217,18 @@ class ProducerImpl : virtual public Producer, virtual public HandleImpl {
                      const std::vector<char> *key,
                      void *msg_opaque);
 
+  ErrorCode produce (const std::string topic_name, int32_t partition,
+                     int msgflags,
+                     void *payload, size_t len,
+                     const void *key, size_t key_len,
+                     int64_t timestamp, void *msg_opaque);
+
   ErrorCode produce (const std::string topic_name, int32_t partition,
                      int msgflags,
                      void *payload, size_t len,
                      const void *key, size_t key_len,
                      int64_t timestamp,
+                     RdKafka::Headers *headers,
                      void *msg_opaque);
 
   ErrorCode flush (int timeout_ms) {
@@ -904,6 +1236,11 @@ class ProducerImpl : virtual public Producer, virtual public HandleImpl {
 								timeout_ms));
   }
 
+  ErrorCode purge (int purge_flags) {
+	  return static_cast<RdKafka::ErrorCode>(rd_kafka_purge(rk_,
+                                                                (int)purge_flags));
+  }
+
   static Producer *create (Conf *conf, std::string &errstr);
 
 };
