@@ -448,7 +448,7 @@ size_t rd_buf_write (rd_buf_t *rbuf, const void *payload, size_t size) {
 
         while (remains > 0) {
                 void *p;
-                rd_segment_t *seg;
+                rd_segment_t *seg = NULL;
                 size_t segremains = rd_buf_get_writable0(rbuf, &seg, &p);
                 size_t wlen = RD_MIN(remains, segremains);
 
@@ -761,9 +761,6 @@ size_t rd_slice_reader0 (rd_slice_t *slice, const void **p, int update_pos) {
         if (unlikely(!seg || seg->seg_absof+rof >= slice->end))
                 return 0;
 
-        rd_assert(seg->seg_absof+rof <= slice->end);
-
-
         *p   = (const void *)(seg->seg_p + rof);
         rlen = RD_MIN(seg->seg_of - rof, rd_slice_remains(slice));
 
@@ -870,6 +867,56 @@ size_t rd_slice_peek (const rd_slice_t *slice, size_t offset,
 }
 
 
+/**
+ * @brief Read a varint-encoded signed integer from \p slice,
+ *        storing the decoded number in \p nump on success (return value > 0).
+ *
+ * @returns the number of bytes read on success or 0 in case of
+ *          buffer underflow.
+ */
+size_t rd_slice_read_varint (rd_slice_t *slice, int64_t *nump) {
+        uint64_t num = 0;
+        int shift = 0;
+        size_t rof = slice->rof;
+        const rd_segment_t *seg;
+
+        /* Traverse segments, byte for byte, until varint is decoded
+         * or no more segments available (underflow). */
+        for (seg = slice->seg ; seg ; seg = TAILQ_NEXT(seg, seg_link)) {
+                for ( ; rof < seg->seg_of ; rof++) {
+                        unsigned char oct;
+
+                        if (unlikely(seg->seg_absof+rof >= slice->end))
+                                return 0; /* Underflow */
+
+                        oct = *(const unsigned char *)(seg->seg_p + rof);
+
+                        num |= (uint64_t)(oct & 0x7f) << shift;
+                        shift += 7;
+
+                        if (!(oct & 0x80)) {
+                                /* Done: no more bytes expected */
+
+                                /* Zig-zag decoding */
+                                *nump = (int64_t)((num >> 1) ^
+                                                  -(int64_t)(num & 1));
+
+                                /* Update slice's read pointer and offset */
+                                if (slice->seg != seg)
+                                        slice->seg = seg;
+                                slice->rof = rof + 1; /* including the +1 byte
+                                                       * that was just read */
+
+                                return shift / 7;
+                        }
+                }
+
+                rof = 0;
+        }
+
+        return 0; /* Underflow */
+}
+
 
 /**
  * @returns a pointer to \p size contiguous bytes at the current read offset.
