@@ -53,6 +53,18 @@ const char *rd_kafka_event_name (const rd_kafka_event_t *rkev) {
 		return "OffsetCommit";
 	case RD_KAFKA_EVENT_STATS:
 		return "Stats";
+        case RD_KAFKA_EVENT_CREATETOPICS_RESULT:
+                return "CreateTopicsResult";
+        case RD_KAFKA_EVENT_DELETETOPICS_RESULT:
+                return "DeleteTopicsResult";
+        case RD_KAFKA_EVENT_CREATEPARTITIONS_RESULT:
+                return "CreatePartitionsResult";
+        case RD_KAFKA_EVENT_ALTERCONFIGS_RESULT:
+                return "AlterConfigsResult";
+        case RD_KAFKA_EVENT_DESCRIBECONFIGS_RESULT:
+                return "DescribeConfigsResult";
+        case RD_KAFKA_EVENT_OAUTHBEARER_TOKEN_REFRESH:
+                return "SaslOAuthBearerTokenRefresh";
 	default:
 		return "?unknown?";
 	}
@@ -134,7 +146,7 @@ size_t rd_kafka_event_message_count (rd_kafka_event_t *rkev) {
 	switch (rkev->rko_evtype)
 	{
 	case RD_KAFKA_EVENT_DR:
-		return rd_atomic32_get(&rkev->rko_u.dr.msgq.rkmq_msg_cnt);
+                return (size_t)rkev->rko_u.dr.msgq.rkmq_msg_cnt;
 	case RD_KAFKA_EVENT_FETCH:
 		return 1;
 	default:
@@ -143,6 +155,18 @@ size_t rd_kafka_event_message_count (rd_kafka_event_t *rkev) {
 }
 
 
+const char *rd_kafka_event_config_string (rd_kafka_event_t *rkev) {
+        switch (rkev->rko_evtype)
+        {
+#if WITH_SASL_OAUTHBEARER
+        case RD_KAFKA_EVENT_OAUTHBEARER_TOKEN_REFRESH:
+                return rkev->rko_rk->rk_conf.sasl.oauthbearer_config;
+#endif
+        default:
+                return NULL;
+        }
+}
+
 rd_kafka_resp_err_t rd_kafka_event_error (rd_kafka_event_t *rkev) {
 	return rkev->rko_err;
 }
@@ -154,10 +178,20 @@ const char *rd_kafka_event_error_string (rd_kafka_event_t *rkev) {
 	case RD_KAFKA_OP_CONSUMER_ERR:
 		if (rkev->rko_u.err.errstr)
 			return rkev->rko_u.err.errstr;
-		/* FALLTHRU */
-	default:
-		return rd_kafka_err2str(rkev->rko_err);
-	}
+                break;
+        case RD_KAFKA_OP_ADMIN_RESULT:
+                if (rkev->rko_u.admin_result.errstr)
+                        return rkev->rko_u.admin_result.errstr;
+                break;
+        default:
+                break;
+        }
+
+        return rd_kafka_err2str(rkev->rko_err);
+}
+
+int rd_kafka_event_error_is_fatal (rd_kafka_event_t *rkev) {
+        return rkev->rko_u.err.fatal;
 }
 
 
@@ -166,6 +200,8 @@ void *rd_kafka_event_opaque (rd_kafka_event_t *rkev) {
 	{
 	case RD_KAFKA_OP_OFFSET_COMMIT:
 		return rkev->rko_u.offset_commit.opaque;
+        case RD_KAFKA_OP_ADMIN_RESULT:
+                return rkev->rko_u.admin_result.opaque;
 	default:
 		return NULL;
 	}
@@ -230,3 +266,49 @@ rd_kafka_event_topic_partition (rd_kafka_event_t *rkev) {
 	return rktpar;
 
 }
+
+
+
+const rd_kafka_CreateTopics_result_t *
+rd_kafka_event_CreateTopics_result (rd_kafka_event_t *rkev) {
+        if (!rkev || rkev->rko_evtype != RD_KAFKA_EVENT_CREATETOPICS_RESULT)
+                return NULL;
+        else
+                return (const rd_kafka_CreateTopics_result_t *)rkev;
+}
+
+
+const rd_kafka_DeleteTopics_result_t *
+rd_kafka_event_DeleteTopics_result (rd_kafka_event_t *rkev) {
+        if (!rkev || rkev->rko_evtype != RD_KAFKA_EVENT_DELETETOPICS_RESULT)
+                return NULL;
+        else
+                return (const rd_kafka_DeleteTopics_result_t *)rkev;
+}
+
+
+const rd_kafka_CreatePartitions_result_t *
+rd_kafka_event_CreatePartitions_result (rd_kafka_event_t *rkev) {
+        if (!rkev || rkev->rko_evtype != RD_KAFKA_EVENT_CREATEPARTITIONS_RESULT)
+                return NULL;
+        else
+                return (const rd_kafka_CreatePartitions_result_t *)rkev;
+}
+
+
+const rd_kafka_AlterConfigs_result_t *
+rd_kafka_event_AlterConfigs_result (rd_kafka_event_t *rkev) {
+        if (!rkev || rkev->rko_evtype != RD_KAFKA_EVENT_ALTERCONFIGS_RESULT)
+                return NULL;
+        else
+                return (const rd_kafka_AlterConfigs_result_t *)rkev;
+}
+
+
+const rd_kafka_DescribeConfigs_result_t *
+rd_kafka_event_DescribeConfigs_result (rd_kafka_event_t *rkev) {
+        if (!rkev || rkev->rko_evtype != RD_KAFKA_EVENT_DESCRIBECONFIGS_RESULT)
+                return NULL;
+        else
+                return (const rd_kafka_DescribeConfigs_result_t *)rkev;
+}
