@@ -45,6 +45,8 @@ typedef struct rd_kafka_interceptor_method_s {
                 rd_kafka_interceptor_f_on_consume_t *on_consume;
                 rd_kafka_interceptor_f_on_commit_t  *on_commit;
                 rd_kafka_interceptor_f_on_request_sent_t *on_request_sent;
+                rd_kafka_interceptor_f_on_thread_start_t *on_thread_start;
+                rd_kafka_interceptor_f_on_thread_exit_t  *on_thread_exit;
                 void *generic; /* For easy assignment */
 
         } u;
@@ -174,6 +176,8 @@ void rd_kafka_interceptors_destroy (rd_kafka_conf_t *conf) {
         rd_list_destroy(&conf->interceptors.on_consume);
         rd_list_destroy(&conf->interceptors.on_commit);
         rd_list_destroy(&conf->interceptors.on_request_sent);
+        rd_list_destroy(&conf->interceptors.on_thread_start);
+        rd_list_destroy(&conf->interceptors.on_thread_exit);
 
         /* Interceptor config */
         rd_list_destroy(&conf->interceptors.config);
@@ -216,6 +220,12 @@ rd_kafka_interceptors_init (rd_kafka_conf_t *conf) {
         rd_list_init(&conf->interceptors.on_request_sent, 0,
                      rd_kafka_interceptor_method_destroy)
                 ->rl_flags |= RD_LIST_F_UNIQUE;
+        rd_list_init(&conf->interceptors.on_thread_start, 0,
+                     rd_kafka_interceptor_method_destroy)
+                ->rl_flags |= RD_LIST_F_UNIQUE;
+        rd_list_init(&conf->interceptors.on_thread_exit, 0,
+                     rd_kafka_interceptor_method_destroy)
+                ->rl_flags |= RD_LIST_F_UNIQUE;
 
         /* Interceptor config */
         rd_list_init(&conf->interceptors.config, 0,
@@ -557,6 +567,48 @@ void rd_kafka_interceptors_on_request_sent (rd_kafka_t *rk,
 }
 
 
+void
+rd_kafka_interceptors_on_thread_start (rd_kafka_t *rk,
+                                       rd_kafka_thread_type_t thread_type) {
+        rd_kafka_interceptor_method_t *method;
+        int i;
+
+        RD_LIST_FOREACH(method, &rk->rk_conf.interceptors.on_thread_start, i) {
+                rd_kafka_resp_err_t ic_err;
+
+                ic_err = method->u.on_thread_start(rk,
+                                                   thread_type,
+                                                   rd_kafka_thread_name,
+                                                   method->ic_opaque);
+                if (unlikely(ic_err))
+                        rd_kafka_interceptor_failed(rk, method,
+                                                    "on_thread_start",
+                                                    ic_err, NULL, NULL);
+        }
+}
+
+
+void rd_kafka_interceptors_on_thread_exit (rd_kafka_t *rk,
+                                           rd_kafka_thread_type_t thread_type) {
+        rd_kafka_interceptor_method_t *method;
+        int i;
+
+        RD_LIST_FOREACH(method, &rk->rk_conf.interceptors.on_thread_exit, i) {
+                rd_kafka_resp_err_t ic_err;
+
+                ic_err = method->u.on_thread_exit(rk,
+                                                  thread_type,
+                                                  rd_kafka_thread_name,
+                                                  method->ic_opaque);
+                if (unlikely(ic_err))
+                        rd_kafka_interceptor_failed(rk, method,
+                                                    "on_thread_exit",
+                                                    ic_err, NULL, NULL);
+        }
+}
+
+
+
 /**
  * @name Public API (backend)
  * @{
@@ -679,3 +731,31 @@ rd_kafka_interceptor_add_on_request_sent (
                                                ic_name, (void *)on_request_sent,
                                                ic_opaque);
 }
+
+
+rd_kafka_resp_err_t
+rd_kafka_interceptor_add_on_thread_start (
+        rd_kafka_t *rk, const char *ic_name,
+        rd_kafka_interceptor_f_on_thread_start_t *on_thread_start,
+        void *ic_opaque) {
+        assert(!rk->rk_initialized);
+        return rd_kafka_interceptor_method_add(&rk->rk_conf.interceptors.
+                                               on_thread_start,
+                                               ic_name,
+                                               (void *)on_thread_start,
+                                               ic_opaque);
+}
+
+
+rd_kafka_resp_err_t
+rd_kafka_interceptor_add_on_thread_exit (
+        rd_kafka_t *rk, const char *ic_name,
+        rd_kafka_interceptor_f_on_thread_exit_t *on_thread_exit,
+        void *ic_opaque) {
+        assert(!rk->rk_initialized);
+        return rd_kafka_interceptor_method_add(&rk->rk_conf.interceptors.
+                                               on_thread_exit,
+                                               ic_name,
+                                               (void *)on_thread_exit,
+                                               ic_opaque);
+}
