@@ -338,31 +338,6 @@ rd_kafka_offset_file_commit (rd_kafka_toppar_t *rktp) {
 }
 
 
-/**
- * Enqueue offset_commit_cb op, if configured.
- *
- */
-void rd_kafka_offset_commit_cb_op (rd_kafka_t *rk,
-				   rd_kafka_resp_err_t err,
-				   const rd_kafka_topic_partition_list_t *offsets) {
-	rd_kafka_op_t *rko;
-
-        if (!(rk->rk_conf.enabled_events & RD_KAFKA_EVENT_OFFSET_COMMIT))
-		return;
-
-	rko = rd_kafka_op_new(RD_KAFKA_OP_OFFSET_COMMIT|RD_KAFKA_OP_REPLY);
-        rd_kafka_op_set_prio(rko, RD_KAFKA_PRIO_HIGH);
-	rko->rko_err = err;
-	rko->rko_u.offset_commit.cb = rk->rk_conf.offset_commit_cb;/*maybe NULL*/
-	rko->rko_u.offset_commit.opaque = rk->rk_conf.opaque;
-	if (offsets)
-		rko->rko_u.offset_commit.partitions =
-			rd_kafka_topic_partition_list_copy(offsets);
-	rd_kafka_q_enq(rk->rk_rep, rko);
-}
-
-
-
 
 /**
  * Commit a list of offsets asynchronously. Response will be queued on 'replyq'.
@@ -421,18 +396,18 @@ rd_kafka_commit (rd_kafka_t *rk,
         if (!(rkcg = rd_kafka_cgrp_get(rk)))
                 return RD_KAFKA_RESP_ERR__UNKNOWN_GROUP;
 
-        if (!async)
+        if (!async) {
                 repq = rd_kafka_q_new(rk);
+                rq = RD_KAFKA_REPLYQ(repq, 0);
+        }
 
-        if (!async) 
-		rq = RD_KAFKA_REPLYQ(repq, 0);
- 
         err = rd_kafka_commit0(rk, offsets, NULL, rq, NULL, NULL, "manual");
 
-        if (!err && !async) {
-		err = rd_kafka_q_wait_result(repq, RD_POLL_INFINITE);
-		rd_kafka_q_destroy(repq);
-        }
+        if (!err && !async)
+                err = rd_kafka_q_wait_result(repq, RD_POLL_INFINITE);
+
+        if (!async)
+                rd_kafka_q_destroy_owner(repq);
 
 	return err;
 }
@@ -503,7 +478,10 @@ rd_kafka_commit_queue (rd_kafka_t *rk,
                         rd_kafka_op_destroy(rko);
                 }
 
-                rd_kafka_q_destroy(rkq);
+                if (rkqu)
+                        rd_kafka_q_destroy(rkq);
+                else
+                        rd_kafka_q_destroy_owner(rkq);
 	}
 
 	return err;
@@ -697,6 +675,9 @@ rd_kafka_offsets_store (rd_kafka_t *rk,
         int i;
         int ok_cnt = 0;
 
+        if (rk->rk_conf.enable_auto_offset_store)
+                return RD_KAFKA_RESP_ERR__INVALID_ARG;
+
         for (i = 0 ; i < offsets->cnt ; i++) {
                 rd_kafka_topic_partition_t *rktpar = &offsets->elems[i];
                 shptr_rd_kafka_toppar_t *s_rktp;
@@ -715,7 +696,7 @@ rd_kafka_offsets_store (rd_kafka_t *rk,
                 ok_cnt++;
         }
 
-        return offsets->cnt > 0 && ok_cnt < offsets->cnt ?
+        return offsets->cnt > 0 && ok_cnt == 0 ?
                 RD_KAFKA_RESP_ERR__UNKNOWN_PARTITION :
                 RD_KAFKA_RESP_ERR_NO_ERROR;
 }
