@@ -56,6 +56,8 @@ struct rd_kafkap_reqhdr {
 #define RD_KAFKAP_Metadata      3
 #define RD_KAFKAP_LeaderAndIsr  4
 #define RD_KAFKAP_StopReplica   5
+#define RD_KAFKAP_UpdateMetadata 6
+#define RD_KAFKAP_ControlledShutdown 7
 #define RD_KAFKAP_OffsetCommit  8
 #define RD_KAFKAP_OffsetFetch   9
 #define RD_KAFKAP_GroupCoordinator 10
@@ -119,6 +121,8 @@ const char *rd_kafka_ApiKey2str (int16_t ApiKey) {
                 [RD_KAFKAP_Metadata] = "Metadata",
                 [RD_KAFKAP_LeaderAndIsr] = "LeaderAndIsr",
                 [RD_KAFKAP_StopReplica] = "StopReplica",
+                [RD_KAFKAP_UpdateMetadata] = "UpdateMetadata",
+                [RD_KAFKAP_ControlledShutdown] = "ControlledShutdown",
                 [RD_KAFKAP_OffsetCommit] = "OffsetCommit",
                 [RD_KAFKAP_OffsetFetch] = "OffsetFetch",
                 [RD_KAFKAP_GroupCoordinator] = "GroupCoordinator",
@@ -186,16 +190,23 @@ struct rd_kafka_ApiVersion {
 /**
  * @brief ApiVersion.ApiKey comparator.
  */
-static RD_UNUSED int rd_kafka_ApiVersion_key_cmp (const void *_a, const void *_b) {
-	const struct rd_kafka_ApiVersion *a = _a, *b = _b;
-
-	return a->ApiKey - b->ApiKey;
+static RD_UNUSED
+int rd_kafka_ApiVersion_key_cmp (const void *_a, const void *_b) {
+        const struct rd_kafka_ApiVersion *a = _a, *b = _b;
+        return RD_CMP(a->ApiKey, b->ApiKey);
 }
 
 
 
-#define RD_KAFKAP_READ_UNCOMMITTED  0
-#define RD_KAFKAP_READ_COMMITTED    1
+typedef enum {
+        RD_KAFKA_READ_UNCOMMITTED = 0,
+        RD_KAFKA_READ_COMMITTED = 1
+} rd_kafka_isolation_level_t;
+
+
+
+#define RD_KAFKA_CTRL_MSG_ABORT 0
+#define RD_KAFKA_CTRL_MSG_COMMIT 1
 
 
 /**
@@ -306,7 +317,7 @@ static RD_INLINE RD_UNUSED int rd_kafkap_str_cmp (const rd_kafkap_str_t *a,
 	if (r)
 		return r;
 	else
-		return a->len - b->len;
+                return RD_CMP(a->len, b->len);
 }
 
 static RD_INLINE RD_UNUSED int rd_kafkap_str_cmp_str (const rd_kafkap_str_t *a,
@@ -317,7 +328,7 @@ static RD_INLINE RD_UNUSED int rd_kafkap_str_cmp_str (const rd_kafkap_str_t *a,
 	if (r)
 		return r;
 	else
-		return a->len - len;
+                return RD_CMP(a->len, len);
 }
 
 static RD_INLINE RD_UNUSED int rd_kafkap_str_cmp_str2 (const char *str,
@@ -328,7 +339,7 @@ static RD_INLINE RD_UNUSED int rd_kafkap_str_cmp_str2 (const char *str,
 	if (r)
 		return r;
 	else
-		return len - b->len;
+                return RD_CMP(len, b->len);
 }
 
 
@@ -430,7 +441,7 @@ static RD_INLINE RD_UNUSED int rd_kafkap_bytes_cmp (const rd_kafkap_bytes_t *a,
 	if (r)
 		return r;
 	else
-		return a->len - b->len;
+                return RD_CMP(a->len, b->len);
 }
 
 static RD_INLINE RD_UNUSED
@@ -441,7 +452,7 @@ int rd_kafkap_bytes_cmp_data (const rd_kafkap_bytes_t *a,
 	if (r)
 		return r;
 	else
-		return a->len - len;
+                return RD_CMP(a->len, len);
 }
 
 
@@ -450,7 +461,7 @@ int rd_kafkap_bytes_cmp_data (const rd_kafkap_bytes_t *a,
 typedef struct rd_kafka_buf_s rd_kafka_buf_t;
 
 
-#define RD_KAFKA_NODENAME_SIZE  128
+#define RD_KAFKA_NODENAME_SIZE  256
 
 
 
@@ -518,6 +529,80 @@ typedef struct rd_kafka_buf_s rd_kafka_buf_t;
 #define RD_KAFKAP_MSGSET_V2_OF_LastOffsetDelta  (8+4+4+1+4+2)
 #define RD_KAFKAP_MSGSET_V2_OF_BaseTimestamp    (8+4+4+1+4+2+4)
 #define RD_KAFKAP_MSGSET_V2_OF_MaxTimestamp     (8+4+4+1+4+2+4+8)
+#define RD_KAFKAP_MSGSET_V2_OF_BaseSequence     (8+4+4+1+4+2+4+8+8+8+2)
 #define RD_KAFKAP_MSGSET_V2_OF_RecordCount      (8+4+4+1+4+2+4+8+8+8+2+4)
 
+
+
+
+/**
+ * @name Producer ID and Epoch for the Idempotent Producer
+ * @{
+ *
+ */
+
+/**
+ * @brief Producer ID and Epoch
+ */
+typedef struct rd_kafka_pid_s {
+        int64_t id;     /**< Producer Id */
+        int16_t epoch;  /**< Producer Epoch */
+} rd_kafka_pid_t;
+
+#define RD_KAFKA_PID_INITIALIZER {-1,-1}
+
+/**
+ * @returns true if \p PID is valid
+ */
+#define rd_kafka_pid_valid(PID) ((PID).id != -1)
+
+/**
+ * @brief Check two pids for equality
+ */
+static RD_UNUSED RD_INLINE int rd_kafka_pid_eq (const rd_kafka_pid_t a,
+                                                const rd_kafka_pid_t b) {
+        return a.id == b.id && a.epoch == b.epoch;
+}
+
+/**
+ * @returns the string representation of a PID in a thread-safe
+ *          static buffer.
+ */
+static RD_UNUSED const char *
+rd_kafka_pid2str (const rd_kafka_pid_t pid) {
+        static RD_TLS char buf[2][64];
+        static RD_TLS int i;
+
+        if (!rd_kafka_pid_valid(pid))
+                return "PID{Invalid}";
+
+        i = (i + 1) % 2;
+
+        rd_snprintf(buf[i], sizeof(buf[i]),
+                    "PID{Id:%"PRId64",Epoch:%hd}", pid.id, pid.epoch);
+
+        return buf[i];
+}
+
+/**
+ * @brief Reset the PID to invalid/init state
+ */
+static RD_UNUSED RD_INLINE void rd_kafka_pid_reset (rd_kafka_pid_t *pid) {
+        pid->id = -1;
+        pid->epoch = -1;
+}
+
+
+/**
+ * @brief Bump the epoch of a valid PID
+ */
+static RD_UNUSED RD_INLINE rd_kafka_pid_t
+rd_kafka_pid_bump (const rd_kafka_pid_t old) {
+        rd_kafka_pid_t new = { old.id, ((int)old.epoch + 1) & (int)INT16_MAX };
+        return new;
+}
+
+/**@}*/
+
+
 #endif /* _RDKAFKA_PROTO_H_ */
