@@ -26,7 +26,8 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 
-#pragma once
+#ifndef _RDLOG_H_
+#define _RDLOG_H_
 
 void rd_hexdump (FILE *fp, const char *name, const void *ptr, size_t len);
 
@@ -35,3 +36,5 @@ void rd_iov_print (const char *what, int iov_idx, const struct iovec *iov,
 struct msghdr;
 void rd_msghdr_print (const char *what, const struct msghdr *msg,
                       int hexdump);
+
+#endif /* _RDLOG_H_ */
