@@ -80,7 +80,7 @@ static void test_single_partition (void) {
         rd_kafka_conf_t *conf;
         rd_kafka_topic_conf_t *topic_conf;
         char msg[128];
-        int msgcnt = test_on_ci ? 1000 : 100000;
+        int msgcnt = test_quick ? 100 : 100000;
         int failcnt = 0;
         int i;
         rd_kafka_message_t *rkmessages;
@@ -115,7 +115,9 @@ static void test_single_partition (void) {
                 rkmessages[i].payload  = rd_strdup(msg);
                 rkmessages[i].len      = strlen(msg);
                 rkmessages[i]._private = msgidp;
-                rkmessages[i].partition = 2;
+                rkmessages[i].partition = 2; /* Will be ignored since
+                                              * RD_KAFKA_MSG_F_PARTITION
+                                              * is not supplied. */
         }
 
         r = rd_kafka_produce_batch(rkt, partition, RD_KAFKA_MSG_F_FREE,
@@ -200,7 +202,7 @@ static void test_partitioner (void) {
         rd_kafka_conf_t *conf;
         rd_kafka_topic_conf_t *topic_conf;
         char msg[128];
-        int msgcnt = test_on_ci ? 1000 : 100000;
+        int msgcnt = test_quick ? 100 : 100000;
         int failcnt = 0;
         int i;
         rd_kafka_message_t *rkmessages;
@@ -315,7 +317,7 @@ static void test_per_message_partition_flag (void) {
         rd_kafka_conf_t *conf;
         rd_kafka_topic_conf_t *topic_conf;
         char msg[128];
-        int msgcnt = 1000;
+        int msgcnt = test_quick ? 100 : 1000;
         int failcnt = 0;
         int i;
         int *rkpartition_counts;
@@ -333,7 +335,7 @@ static void test_per_message_partition_flag (void) {
         TEST_SAY("test_per_message_partition_flag: Created kafka instance %s\n",
                  rd_kafka_name(rk));
         topic_name = test_mk_topic_name("0011_per_message_flag", 1);
-        test_create_topic(topic_name, topic_num_partitions, 1);
+        test_create_topic(rk, topic_name, topic_num_partitions, 1);
 
         rkt = rd_kafka_topic_new(rk, topic_name,
                                  topic_conf);
@@ -449,7 +451,7 @@ static void test_message_partitioner_wo_per_message_flag (void) {
         rd_kafka_conf_t *conf;
         rd_kafka_topic_conf_t *topic_conf;
         char msg[128];
-        int msgcnt = 1000;
+        int msgcnt = test_quick ? 100 : 1000;
         int failcnt = 0;
         int i;
         rd_kafka_message_t *rkmessages;
