@@ -47,7 +47,7 @@ static int do_test_consume_batch (void) {
         rd_kafka_queue_t *rkq;
         rd_kafka_topic_t *rkts[topic_cnt];
 	rd_kafka_resp_err_t err;
-        const int msgcnt = 10000;
+        const int msgcnt = test_quick ? 1000 : 10000;
         uint64_t testid;
         int i, p;
         int batch_cnt = 0;
@@ -111,7 +111,7 @@ static int do_test_consume_batch (void) {
 
                 if (r == -1)
                         TEST_FAIL("Failed to consume messages: %s\n",
-                                  rd_kafka_err2str(rd_kafka_errno2err(errno)));
+                                  rd_kafka_err2str(rd_kafka_last_error()));
 
                 remains -= (int)r;
 
