@@ -7,11 +7,11 @@ CYAN='\033[36m'
 CCLR='\033[0m'
 
 if [ -z "$1" ]; then
-    echo "Usage: $0 [-..] <execfile> [modes..]"
+    echo "Usage: $0 [-..] [modes..]"
     echo ""
-    echo "  Modes: bare valgrind helgrind drd gdb"
+    echo "  Modes: bare valgrind helgrind drd gdb lldb bash"
     echo "  Options:"
-    echo "   -..    - Command arguments (pass thru)"
+    echo "   -..    - test-runner command arguments (pass thru)"
     exit 1
 fi
 
@@ -22,7 +22,8 @@ while [[ $1 == -* ]]; do
     shift
 done
 
-TEST=$1
+TEST=./test-runner
+
 if [ ! -z "$2" ]; then
     MODES=$2
 else
@@ -77,13 +78,25 @@ for mode in $MODES; do
 	    RET=$?
 	    ;;
         gdb)
-            gdb $ARGS $TEST
+            if [[ -f gdb.run ]]; then
+                gdb -x gdb.run $ARGS $TEST
+            else
+                gdb $ARGS $TEST
+            fi
             RET=$?
             ;;
 	bare)
 	    $TEST $ARGS
 	    RET=$?
 	    ;;
+        lldb)
+            lldb -b -o "process launch --environment DYLD_LIBRARY_PATH=$DYLD_LIBRARY_PATH" -- $TEST $ARGS
+            RET=$?
+            ;;
+	bash)
+	    PS1="[run-test.sh] $PS1" bash
+	    RET=$?
+	    ;;
 	*)
 	    echo -e "${RED}### Unknown mode $mode for $TEST ###${CCLR}"
 	    RET=1
