@@ -48,13 +48,22 @@ while true ; do
             else
                 export TESTS=$t
             fi
-            ./run-test.sh $ARGS ./merged $mode || (echo "Failed on iteration $iter, test $t, mode $mode" ; exit 1)
+            ./run-test.sh $ARGS $mode || (echo "Failed on iteration $iter, test $t, mode $mode" ; exit 1)
         done
     done
 
 
     if [[ "$DELETE_TOPICS" == "y" ]]; then
-	./delete-test-topics.sh $ZK_ADDRESS ~/src/kafka/bin/kafka-topics.sh || true
+        # Delete topics using Admin API, which is very fast
+        # leads to sub-sequent test failures because of the background
+        # deletes in Kafka still taking a long time:
+        #
+        #make delete_topics
+
+        # Delete topic-by-topic using kafka-topics for each one,
+        # very slow but topics are properly deleted before the script
+        # returns.
+        ./delete-test-topics.sh $ZK_ADDRESS ~/src/kafka/bin/kafka-topics.sh || true
     fi
 done
 
